/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */
/** \file ortp_recv.c paraslash's ortp receiver */

#include <ortp/ortp.h>
#include "para.h"

#include "ortp.h"
#include "recv.h"
#include "ortp_recv.cmdline.h"

#include "error.h"
#include "audiod.h"
#include "string.h"

#define CHUNK_SIZE 128 * 1024

extern int msg_to_buf(mblk_t *, char *, int);

/**
 * data specific to the ortp receiver
 *
 * \sa receiver receiver_node
 */
struct private_ortp_recv_data {

/**
 *
 *
 * whether a header was received
 *
 * A flag indicating whether the ortp receiver already received a packet which
 * contains the audio file header.
 *
 * This flag has no effect if the audio stream indicates that no extra headers
 * will be sent (mp3).  Otherwise, all data packets are dropped until the
 * header is received.
 */
int have_header;
/** the ortp session this instance of the receiver belongs to */
RtpSession *session;
/** the time the first data or header packet was received */
struct timeval start;
/** ETA of the next chunk */
struct timeval next_chunk;
/** number of consecutive bad chunks */
int c_bad;
/** the current timestamp which is passed to the receiving function of libortp */
uint32_t timestamp;
/** \a timestamp increases by this amount */
uint32_t chunk_ts;
};

static int ortp_recv_pre_select(struct receiver_node *rn,
		__a_unused fd_set *rfds, __a_unused fd_set *wfds,
		struct timeval *timeout)
{
	struct private_ortp_recv_data *pord = rn->private_data;
	struct timeval now, tmp;

	gettimeofday(&now, NULL);
	if (tv_diff(&now, &pord->next_chunk, &tmp) >= 0) {
		tmp.tv_sec = 0;
		tmp.tv_usec = 1000;
	}
	if (tv_diff(timeout, &tmp, NULL) > 0)
		*timeout = tmp;
	return -1; /* we did not modify the fd sets */
}

static void compute_next_chunk(unsigned chunk_time,
		struct private_ortp_recv_data *pord)
{
	struct timeval chunk_tv = {0, chunk_time};
	struct timeval tmp;

	tv_add(&chunk_tv, &pord->next_chunk, &tmp);
	pord->next_chunk = tmp;
	pord->timestamp += pord->chunk_ts;
	PARA_DEBUG_LOG("next chunk (ts = %d) due at %lu:%lu\n",
		pord->timestamp, pord->next_chunk.tv_sec,
		pord->next_chunk.tv_usec);
}

static int ortp_recv_post_select(struct receiver_node *rn,
		__a_unused int select_ret, __a_unused fd_set *rfds,
		__a_unused fd_set *wfds)
{
	struct private_ortp_recv_data *pord = rn->private_data;
	mblk_t *mp;
	int ret, packet_type, stream_type;
	char tmpbuf[CHUNK_SIZE + 3];
	struct timeval now;
	unsigned chunk_time;

	gettimeofday(&now, NULL);
//	PARA_DEBUG_LOG("rn: %p, pord: %p, session: %p\n", rn, pord, pord->session);
	if (pord->start.tv_sec) {
		struct timeval diff;
		if (tv_diff(&now, &pord->next_chunk, &diff) < 0)
				return 1;
	}
	mp = rtp_session_recvm_with_ts(pord->session, pord->timestamp);
	if (!mp) {
		struct timeval min_delay = {0, 100};
//		PARA_INFO_LOG("nope, chunk_ts = %d, loaded: %d, bad: %d\n",
//			pord->timestamp, rn->loaded, pord->c_bad);
		pord->c_bad++;
		if ((pord->c_bad > 5000 && pord->start.tv_sec) || pord->c_bad > 10000)
			return -E_TOO_MANY_BAD_CHUNKS;
		tv_add(&now, &min_delay, &pord->next_chunk);
		return 1;
	}
	/* okay, we have a chunk of data */
	if (!pord->start.tv_sec)
		pord->start = now;
	ret = msg_to_buf(mp, tmpbuf, CHUNK_SIZE);
//	PARA_DEBUG_LOG("have it ts = %d, chunk_ts = %d, loaded: %d, "
//		"bad: %d, len: %d\n", pord->timestamp, pord->chunk_ts,
//		rn->loaded, pord->c_bad, ret);
	if (ret < ORTP_AUDIO_HEADER_LEN) {
		if (ret < 0)
			ret = -E_MSG_TO_BUF;
		else
			ret = 0;
		goto err_out;
	}
	packet_type = READ_PACKET_TYPE(tmpbuf);
	stream_type = READ_STREAM_TYPE(tmpbuf);
	chunk_time = READ_CHUNK_TIME(tmpbuf);
	pord->chunk_ts = READ_CHUNK_TS(tmpbuf);
//	PARA_INFO_LOG("packet type: %d, stream type: %d, chunk time: %u,"
//		" chunk_ts: %u, loaded: %u, bad: %d\n", packet_type,
//		stream_type, chunk_time, pord->chunk_ts, rn->loaded, pord->c_bad);
	switch (packet_type) {
	unsigned header_len, payload_len;
	case ORTP_EOF:
		ret = 0;
		goto err_out;
	case ORTP_BOF:
		PARA_INFO_LOG("bof (%d)\n", ret);
		pord->have_header = 1;
		/* fall through */
	case ORTP_DATA:
		if (!pord->have_header && stream_type)
		/* can't use the data, wait for header */
			goto success;
		if (ret + rn->loaded >= CHUNK_SIZE + ORTP_AUDIO_HEADER_LEN) {
			ret = -E_OVERRUN;
			goto err_out;
		}
		if (ret > ORTP_AUDIO_HEADER_LEN) {
			memcpy(rn->buf + rn->loaded, tmpbuf + ORTP_AUDIO_HEADER_LEN,
				ret - ORTP_AUDIO_HEADER_LEN);
			rn->loaded += ret - ORTP_AUDIO_HEADER_LEN;
		}
		goto success;
	case ORTP_HEADER:
		header_len = READ_HEADER_LEN(tmpbuf);
		PARA_DEBUG_LOG("header packet (%d bytes), header len: %d\n",
			ret, header_len);
		if (!pord->have_header) {
			pord->have_header = 1;
			memcpy(rn->buf, tmpbuf + ORTP_AUDIO_HEADER_LEN,
				ret - ORTP_AUDIO_HEADER_LEN);
			rn->loaded = ret - ORTP_AUDIO_HEADER_LEN;
			goto success;
		}
		if (header_len + ORTP_AUDIO_HEADER_LEN > ret) {
			ret = -E_INVALID_HEADER;
			goto err_out;
		}
		payload_len = ret - ORTP_AUDIO_HEADER_LEN - header_len;
//		PARA_INFO_LOG("len: %d header_len: %d, payload_len: %d, loaded: %d\n", ret,
//			header_len, payload_len, rn->loaded);
		if (rn->loaded + payload_len > CHUNK_SIZE) {
			ret = -E_OVERRUN;
			goto err_out;
		}
		if (payload_len)
			memcpy(rn->buf + rn->loaded, tmpbuf
				+ (ret - payload_len), payload_len);
		rn->loaded += payload_len;
		goto success;
	}
success:
	freemsg(mp);
	if (pord->c_bad) {
		pord->c_bad = 0;
		pord->next_chunk = now;
	}
	compute_next_chunk(chunk_time, pord);
	return 1;
err_out:
	freemsg(mp);
	return ret;
}

static void ortp_shutdown(void)
{
//	ortp_global_stats_display();
	ortp_exit();
}

static void ortp_recv_close(struct receiver_node *rn)
{
	struct private_ortp_recv_data *pord = rn->private_data;

	rtp_session_destroy(pord->session);
	free(rn->private_data);
	free(rn->buf);
}

static void *ortp_recv_parse_config(int argc, char **argv)
{
	int ret;

	struct ortp_recv_args_info *tmp = para_calloc(sizeof(struct ortp_recv_args_info));

	ret = ortp_recv_cmdline_parser(argc, argv, tmp)? -E_ORTP_SYNTAX : 1;
	if (ret > 0)
		return tmp;
	free(tmp);
	return NULL;
}

static int ortp_recv_open(struct receiver_node *rn)
{
	struct private_ortp_recv_data *pord;
	struct ortp_recv_args_info *conf = rn->conf;

	rn->buf = para_calloc(CHUNK_SIZE);

	rn->private_data = para_calloc(sizeof(struct private_ortp_recv_data));
	pord = rn->private_data;
	pord->session = rtp_session_new(RTP_SESSION_RECVONLY);
	PARA_NOTICE_LOG("receiving from %s:%d\n", conf->host_arg, conf->port_arg);
	rtp_session_set_local_addr(pord->session, conf->host_arg, conf->port_arg);
	rtp_session_set_payload_type(pord->session, PAYLOAD_AUDIO_CONTINUOUS);
	if (conf->jitter_compensation_arg) {
		rtp_session_enable_adaptive_jitter_compensation(pord->session, TRUE);
		rtp_session_set_jitter_compensation(pord->session,
			conf->jitter_compensation_arg);
	}
	return 1;
}

/**
 * the init function of the ortp receiver
 *
 * \param r pointer to the receiver struct to initialize
 *
 * Initialize all function pointers of \a r and call libortp's ortp_init().
 */
void ortp_recv_init(struct receiver *r)
{
	r->shutdown = ortp_shutdown;
	r->open = ortp_recv_open;
	r->close = ortp_recv_close;
	r->pre_select = ortp_recv_pre_select;
	r->post_select = ortp_recv_post_select;
	r->parse_config = ortp_recv_parse_config;

	ortp_init();
}
